import io
import logging
from typing import Awaitable, Dict, List, Optional, TypeVar
import typing
import discord
from discord.channel import VoiceChannel
import discord.ext.commands
import asyncio
from discord.member import VoiceState
from discord.user import User
import librosa
import soundfile
import wave

from freem_bots.asyncio_queue_log_handler import AsyncioQueueLogHandler
from freem_bots.configuration import Config
from freem_bots.random_provider import RandomProvider
from freem_bots.tts import TTS, TTSPause, TTSVoice, TTSVoiceline, TTSVoicelinePart, TTSVoicelineSpeed


class DiscordBot(discord.ext.commands.Bot):  # type: ignore
	_tvC = TypeVar('_tvC', bound = Config)

	def __init__(self, configuration: _tvC) -> None:
		self._log_queue: 'asyncio.Queue[str]' = asyncio.Queue()
		self._logger = logging.getLogger('bot-discord')
		handler = AsyncioQueueLogHandler(self._log_queue)
		logging.root.addHandler(handler)
		super().__init__(command_prefix = '/', intents = self._intents)
		self._configuration = configuration
		self._random_provider: RandomProvider = RandomProvider()
		self.__load_opus()
		self._tts = TTS(self._configuration)
		self._init_tasks: List[Awaitable[None]] = [self.__task_log_to_log_channel()]
		self._voice_lock = asyncio.Lock()
		self._user_channels: Dict[int, VoiceChannel] = {}

	@property
	def _intents(self) -> discord.Intents:
		intents = discord.Intents.all()
		return intents

	async def on_ready(self) -> None:
		''' Invoked when bot connects to Discord servers '''
		self._logger.info('Bot connected under %s', self.user.name)
		for waiting_task in self._init_tasks:
			self.loop.create_task(waiting_task)

	async def on_voice_state_update(  # pylint: disable=unused-argument
		self,
		member: discord.Member,
		before: VoiceState,
		after: VoiceState,
	) -> None:
		if after.channel is None:
			self._logger.debug('Registered voice state update, disconnection for user %s', member.name)
			self._user_channels.pop(member.id)
		elif after.channel is not None and isinstance(after.channel, VoiceChannel):
			self._logger.debug('Registered voice state update, user %s in channel %s', member.name, after.channel.name)
			self._user_channels[member.id] = after.channel

	async def get_audio_for_text(self, text: str, voice: TTSVoice, username: str) -> bytes:
		''' Returns PCM for given lines '''
		username_lines = text.split('{}')
		message_voicelines: List[TTSVoicelinePart] = []
		for username_line in username_lines:
			message_voicelines.append(TTSVoicelinePart(voice, username_line.strip(), TTSVoicelineSpeed.NORMAL))
			message_voicelines.append(
				TTSVoicelinePart(voice, username.strip(), TTSVoicelineSpeed.SLOWER, prepended_pause = TTSPause.SHORT)
			)
		message_voicelines.pop()
		voiceline = TTSVoiceline(message_voicelines)
		return await self.get_voiceline_pcm(voiceline)

	async def get_audio_for_text_simple(self, text: str, voice: TTSVoice) -> bytes:
		''' Gets PCM for a simple string '''
		voiceline = TTSVoiceline([TTSVoicelinePart(voice, text)])
		return await self.get_voiceline_pcm(voiceline)

	async def get_voiceline_pcm(self, voiceline: TTSVoiceline) -> bytes:
		''' Gets PCM for voiceline content '''
		sample_rate = self.redis.get_int('bot_audio_processing_target_sample_rate', 96000)
		byts = await self._tts.get_audio_bytes(voiceline, sample_rate)
		return DiscordBot.get_pcm_from_bytes(byts)

	async def play_audiofile(self, filename: str, voice_channel: VoiceChannel) -> None:
		if '.' in filename:
			self._logger.warning('Refusing to resolve insecure path: filename=%s', filename)
			return
		full_path = f'audio_files/{filename}.wav'
		pcms = [self.get_pcm_from_file(full_path)]
		await self.play_pcms_in_voice_channel(voice_channel, pcms)

	def get_pcm_from_file(self, file: str) -> bytes:
		''' Gets PCM from a file '''
		(audio, original_sample_rate) = soundfile.read(file)
		target_samplerate = self.redis.get_int('bot_audio_processing_target_sample_rate', 96000)
		audio = librosa.resample(audio, orig_sr = original_sample_rate, target_sr = target_samplerate)
		soundfile.write(file, audio, target_samplerate)

		audio_file = wave.open(file, 'rb')
		num_frames: int = audio_file.getnframes()
		audio_raw: bytes = audio_file.readframes(num_frames)
		audio_file.close()

		return audio_raw

	@staticmethod
	def get_pcm_from_bytes(sound_bytes: bytes) -> bytes:
		''' Gets PCM from bytes '''
		with io.BytesIO(sound_bytes) as file:
			file.name = 'audio.wav'
			audio_file = wave.open(file, 'rb')
			num_frames: int = audio_file.getnframes()
			audio_raw: bytes = audio_file.readframes(num_frames)
			audio_file.close()

		return audio_raw

	def send_log_to_log_channel(self, message: str) -> None:
		self._log_queue.put_nowait(message)

	def locate_user_in_voice_channel(self, target_user: User) -> Optional[VoiceChannel]:
		if target_user.id in self._user_channels:
			return self._user_channels[target_user.id]
		else:
			return None

	async def play_voiceline_in_voice_channel(
		self,
		voice_channel: VoiceChannel,
		voiceline: TTSVoiceline,
	) -> None:
		byts = await self._tts.get_audio_bytes(voiceline)
		pcms = [DiscordBot.get_pcm_from_bytes(byts)]
		await self.play_pcms_in_voice_channel(voice_channel, pcms)

	async def play_pcms_in_voice_channel(self, voice_channel: VoiceChannel, pcms: List[bytes]) -> None:
		''' Plays a collection of PCMs in a voice channel '''
		# else, remove self, connect to voice channel, play audio, disconnect
		try:
			async with self._voice_lock:
				voice_client: discord.VoiceClient = await voice_channel.connect()
				try:
					pcm = bytes()
					for pcm_ in pcms:
						pcm += pcm_
					pcm_bytes = bytes(pcm)
					audio_stream = io.BytesIO(pcm_bytes)
					audio_source = discord.PCMAudio(audio_stream)
					self._logger.debug('Starting playing')
					voice_client.play(audio_source)
					while voice_client.is_playing():
						await asyncio.sleep(0.1)
					self._logger.debug('Done playing')
					await voice_client.disconnect()
				except Exception:
					await voice_client.disconnect()
					await asyncio.sleep(0.5)
					raise
		except Exception as e:  # pylint: disable=broad-except
			self._logger.error('Invalid state change in voice activity: %s', e)

	async def sync_commands(self) -> None:
		return typing.cast(None, await super().sync_commands())

	async def _get_user_by_uid(self, guild: str, user_id: int) -> discord.Member:
		try:
			target_user_id = user_id
			# find the guild
			target_guilds = [i for i in self.guilds if i.name.lower() == guild.lower()]
			if len(target_guilds) == 0:
				self._logger.fatal('Unable to find %s guild', guild)
				return None
			target_guild = target_guilds[0]
			# get user
			members: List[discord.Member] = [member for member in target_guild.members if member.id == target_user_id]
			if len(members) == 0:
				self._logger.error('Unable to find user in %s', guild)
				return None

			target_user = members[0]
			return target_user
		except discord.DiscordException:
			return None

	def __load_opus(self) -> None:
		if not discord.opus.is_loaded():
			discord.opus._load_default()  # pylint: disable=protected-access
		if not discord.opus.is_loaded():
			raise Exception('Unable to load OPUS audio library')

	async def __task_log_to_log_channel(self) -> None:
		log_channel_id = self._configuration.log_channel_id
		if log_channel_id is None:
			self._logger.warning("Won't log to discord, no log channel configured")
			while not self.is_closed():
				await self._log_queue.get()
				await asyncio.sleep(0.05)
			return
		log_channel = self.get_channel(log_channel_id)
		while not self.is_closed():
			while not self._log_queue.empty():
				message: str = await self._log_queue.get()
				try:
					message = message.replace('`', '\\`')
					message = f'`{message}`'
					await log_channel.send(message)
				except discord.DiscordException:
					await asyncio.sleep(5.0)  # probably hit rate limiting, if discord disconnect, loop will stop anyway
			await asyncio.sleep(0.05)
